//
// This flag can be used to include unsupported or unrecognized sockets
// in the fd tables. It's useful to debug close() leaks
//
#define INCLUDE_UNKNOWN_SOCKET_FDS

#define USE_ZLIB

#define SCAP_NODRIVER_MAX_FD_LOOKUP 20